package io.github.resilience4j.reactor;

import java.io.IOException;
import java.time.Duration;
import java.time.temporal.ChronoUnit;

import org.junit.Test;

import io.github.resilience4j.bulkhead.Bulkhead;
import io.github.resilience4j.bulkhead.BulkheadConfig;
import io.github.resilience4j.circuitbreaker.CircuitBreaker;
import io.github.resilience4j.circuitbreaker.CircuitBreakerConfig;
import io.github.resilience4j.circuitbreaker.CircuitBreakerOpenException;
import io.github.resilience4j.ratelimiter.RateLimiter;
import io.github.resilience4j.ratelimiter.RateLimiterConfig;
import io.github.resilience4j.reactor.bulkhead.operator.BulkheadOperator;
import io.github.resilience4j.reactor.circuitbreaker.operator.CircuitBreakerOperator;
import io.github.resilience4j.reactor.ratelimiter.operator.RateLimiterOperator;
import io.github.resilience4j.reactor.retry.RetryOperator;
import io.github.resilience4j.retry.Retry;
import io.github.resilience4j.retry.RetryConfig;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.core.scheduler.Schedulers;
import reactor.test.StepVerifier;

public class CombinedOperatorsTest {

    private final RateLimiter rateLimiter = RateLimiter.of("test",
            RateLimiterConfig.custom().limitForPeriod(5).timeoutDuration(Duration.ZERO).limitRefreshPeriod(Duration.ofSeconds(10)).build());

    private final CircuitBreaker circuitBreaker = CircuitBreaker.of("test",
            CircuitBreakerConfig.custom()
                    .waitDurationInOpenState(Duration.of(10, ChronoUnit.SECONDS))
                    .ringBufferSizeInClosedState(4)
                    .ringBufferSizeInHalfOpenState(4)
                    .build());

    private Bulkhead bulkhead = Bulkhead
            .of("test", BulkheadConfig.custom().maxConcurrentCalls(1).maxWaitTime(0).build());

    private final RetryConfig config = io.github.resilience4j.retry.RetryConfig.ofDefaults();
    private final Retry retry = Retry.of("testName", config);
    private final RetryOperator<String> retryOperator = RetryOperator.of(retry);

    @Test
    public void shouldEmitEvents() {
        StepVerifier.create(
                Flux.just("Event 1", "Event 2")
                        .transform(BulkheadOperator.of(bulkhead))
                        .transform(RateLimiterOperator.of(rateLimiter))
                        .transform(CircuitBreakerOperator.of(circuitBreaker))
        ).expectNext("Event 1")
                .expectNext("Event 2")
                .verifyComplete();
    }

    @Test
    public void shouldEmitEventsWithRetry() {
        StepVerifier.create(
                Flux.just("Event 1", "Event 2")
                        .transform(retryOperator)
                        .transform(BulkheadOperator.of(bulkhead))
                        .transform(RateLimiterOperator.of(rateLimiter))
                        .transform(CircuitBreakerOperator.of(circuitBreaker))
        ).expectNext("Event 1")
                .expectNext("Event 2")
                .verifyComplete();
    }

    @Test
    public void shouldEmitEvent() {
        StepVerifier.create(
                Mono.just("Event 1")
                        .transform(BulkheadOperator.of(bulkhead))
                        .transform(RateLimiterOperator.of(rateLimiter))
                        .transform(CircuitBreakerOperator.of(circuitBreaker))
        ).expectNext("Event 1")
                .verifyComplete();
    }

    @Test
    public void shouldEmitEventWithRetry() {
        StepVerifier.create(
                Mono.just("Event 1")
                        .transform(retryOperator)
                        .transform(BulkheadOperator.of(bulkhead))
                        .transform(RateLimiterOperator.of(rateLimiter))
                        .transform(CircuitBreakerOperator.of(circuitBreaker))
        ).expectNext("Event 1")
                .verifyComplete();
    }

    @Test
    public void shouldPropagateError() {
        StepVerifier.create(
                Flux.error(new IOException("BAM!"))
                        .transform(BulkheadOperator.of(bulkhead))
                        .transform(RateLimiterOperator.of(rateLimiter))
                        .transform(CircuitBreakerOperator.of(circuitBreaker))
        ).expectError(IOException.class)
                .verify(Duration.ofSeconds(1));
    }

    @Test
    public void shouldEmitErrorWithCircuitBreakerOpenExceptionEvenWhenErrorDuringSubscribe() {
        circuitBreaker.transitionToOpenState();
        StepVerifier.create(
                Flux.error(new IOException("BAM!"))
                        .transform(CircuitBreakerOperator.of(circuitBreaker))
                        .transform(BulkheadOperator.of(bulkhead, Schedulers.immediate()))
                        .transform(RateLimiterOperator.of(rateLimiter, Schedulers.immediate()))
        ).expectError(CircuitBreakerOpenException.class)
                .verify(Duration.ofSeconds(1));
    }

    @Test
    public void shouldEmitErrorWithCircuitBreakerOpenExceptionEvenWhenErrorNotOnSubscribe() {
        circuitBreaker.transitionToOpenState();
        StepVerifier.create(
                Flux.error(new IOException("BAM!"), true)
                        .transform(CircuitBreakerOperator.of(circuitBreaker))
                        .transform(BulkheadOperator.of(bulkhead, Schedulers.immediate()))
                        .transform(RateLimiterOperator.of(rateLimiter, Schedulers.immediate()))
        ).expectError(CircuitBreakerOpenException.class)
                .verify(Duration.ofSeconds(1));
    }
}
